package jdk.net;

import java.net.MulticastSocket;
import java.util.Collections;
import java.net.StandardSocketOptions;
import java.util.HashSet;
import java.net.ServerSocket;
import java.net.Socket;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.DatagramSocket;
import java.security.AccessController;
import java.lang.reflect.Method;
import java.net.SocketOption;
import java.util.Set;
import java.util.HashMap;
import jdk.Exported;

@Exported
public class Sockets
{
    private static final HashMap<Class<?>, Set<SocketOption<?>>> options;
    private static Method siSetOption;
    private static Method siGetOption;
    private static Method dsiSetOption;
    private static Method dsiGetOption;
    
    static {
        options = new HashMap<>();
        initOptionSets();
        AccessController.doPrivileged(new Sockets.Sockets$1());
    }
    
    private static void initMethods() {
        try {
            Class<?> forName = Class.forName("java.net.SocketSecrets");
            (Sockets.siSetOption = forName.getDeclaredMethod("setOption", new Class[] { Object.class, SocketOption.class, Object.class })).setAccessible(true);
            (Sockets.siGetOption = forName.getDeclaredMethod("getOption", new Class[] { Object.class, SocketOption.class })).setAccessible(true);
            (Sockets.dsiSetOption = forName.getDeclaredMethod("setOption", new Class[] { DatagramSocket.class, SocketOption.class, Object.class })).setAccessible(true);
            (Sockets.dsiGetOption = forName.getDeclaredMethod("getOption", new Class[] { DatagramSocket.class, SocketOption.class })).setAccessible(true);
        }
        catch (ReflectiveOperationException ex) {
            throw new InternalError(ex);
        }
    }
    
    private static <T> void invokeSet(Method method, Object o, SocketOption<T> socketOption, T t) throws IOException {
        try {
            method.invoke(null, new Object[] { o, socketOption, t });
        }
        catch (Exception ex) {
            if (ex instanceof InvocationTargetException) {
                Throwable targetException = ((InvocationTargetException)ex).getTargetException();
                if (targetException instanceof IOException) {
                    throw (IOException)targetException;
                }
                if (targetException instanceof RuntimeException) {
                    throw (RuntimeException)targetException;
                }
            }
            throw new RuntimeException(ex);
        }
    }
    
    private static <T> T invokeGet(Method method, Object o, SocketOption<T> socketOption) throws IOException {
        throw new Error("Unresolved compilation problem: \n\tType mismatch: cannot convert from Object to T\n");
    }
    
    private Sockets() {
    }
    
    public static <T> void setOption(Socket socket, SocketOption<T> socketOption, T t) throws IOException {
        if (!isSupported(Socket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        invokeSet(Sockets.siSetOption, socket, socketOption, t);
    }
    
    public static <T> T getOption(Socket socket, SocketOption<T> socketOption) throws IOException {
        if (!isSupported(Socket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        return invokeGet(Sockets.siGetOption, socket, socketOption);
    }
    
    public static <T> void setOption(ServerSocket serverSocket, SocketOption<T> socketOption, T t) throws IOException {
        if (!isSupported(ServerSocket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        invokeSet(Sockets.siSetOption, serverSocket, socketOption, t);
    }
    
    public static <T> T getOption(ServerSocket serverSocket, SocketOption<T> socketOption) throws IOException {
        if (!isSupported(ServerSocket.class, socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        return invokeGet(Sockets.siGetOption, serverSocket, socketOption);
    }
    
    public static <T> void setOption(DatagramSocket datagramSocket, SocketOption<T> socketOption, T t) throws IOException {
        if (!isSupported(datagramSocket.getClass(), socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        invokeSet(Sockets.dsiSetOption, datagramSocket, socketOption, t);
    }
    
    public static <T> T getOption(DatagramSocket datagramSocket, SocketOption<T> socketOption) throws IOException {
        if (!isSupported(datagramSocket.getClass(), socketOption)) {
            throw new UnsupportedOperationException(socketOption.name());
        }
        return invokeGet(Sockets.dsiGetOption, datagramSocket, socketOption);
    }
    
    public static Set<SocketOption<?>> supportedOptions(Class<?> clazz) {
        Set<SocketOption<?>> set = (Set<SocketOption<?>>)Sockets.options.get(clazz);
        if (set == null) {
            throw new IllegalArgumentException("unknown socket type");
        }
        return (Set)set;
    }
    
    private static boolean isSupported(Class<?> clazz, SocketOption<?> socketOption) {
        return supportedOptions(clazz).contains(socketOption);
    }
    
    private static void initOptionSets() {
        boolean flowSupported = ExtendedSocketOptions.PlatformSocketOptions.get().flowSupported();
        boolean keepAliveOptionsSupported = ExtendedSocketOptions.PlatformSocketOptions.get().keepAliveOptionsSupported();
        HashSet<ExtendedSocketOptions> set = new HashSet<>();
        set.add(StandardSocketOptions.SO_KEEPALIVE);
        set.add(StandardSocketOptions.SO_SNDBUF);
        set.add(StandardSocketOptions.SO_RCVBUF);
        set.add(StandardSocketOptions.SO_REUSEADDR);
        set.add(StandardSocketOptions.SO_LINGER);
        set.add(StandardSocketOptions.IP_TOS);
        set.add(StandardSocketOptions.TCP_NODELAY);
        if (flowSupported) {
            set.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        if (keepAliveOptionsSupported) {
            set.add(ExtendedSocketOptions.TCP_KEEPCOUNT);
            set.add(ExtendedSocketOptions.TCP_KEEPIDLE);
            set.add(ExtendedSocketOptions.TCP_KEEPINTERVAL);
        }
        Sockets.options.put(Socket.class, Collections.unmodifiableSet(set));
        HashSet<ExtendedSocketOptions> set2 = new HashSet<>();
        set2.add(StandardSocketOptions.SO_RCVBUF);
        set2.add(StandardSocketOptions.SO_REUSEADDR);
        set2.add(StandardSocketOptions.IP_TOS);
        if (keepAliveOptionsSupported) {
            set2.add(ExtendedSocketOptions.TCP_KEEPCOUNT);
            set2.add(ExtendedSocketOptions.TCP_KEEPIDLE);
            set2.add(ExtendedSocketOptions.TCP_KEEPINTERVAL);
        }
        Sockets.options.put(ServerSocket.class, Collections.unmodifiableSet(set2));
        HashSet<ExtendedSocketOptions> set3 = new HashSet<>();
        set3.add(StandardSocketOptions.SO_SNDBUF);
        set3.add(StandardSocketOptions.SO_RCVBUF);
        set3.add(StandardSocketOptions.SO_REUSEADDR);
        set3.add(StandardSocketOptions.IP_TOS);
        if (flowSupported) {
            set3.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        Sockets.options.put(DatagramSocket.class, Collections.unmodifiableSet(set3));
        HashSet<ExtendedSocketOptions> set4 = new HashSet<>();
        set4.add(StandardSocketOptions.SO_SNDBUF);
        set4.add(StandardSocketOptions.SO_RCVBUF);
        set4.add(StandardSocketOptions.SO_REUSEADDR);
        set4.add(StandardSocketOptions.IP_TOS);
        set4.add(StandardSocketOptions.IP_MULTICAST_IF);
        set4.add(StandardSocketOptions.IP_MULTICAST_TTL);
        set4.add(StandardSocketOptions.IP_MULTICAST_LOOP);
        if (flowSupported) {
            set4.add(ExtendedSocketOptions.SO_FLOW_SLA);
        }
        Sockets.options.put(MulticastSocket.class, Collections.unmodifiableSet(set4));
    }
}
